/*
 * scout-robotd/console.c
 *
 * this file is responsible for the network console interface to the
 * robot.  this interface allows the user to connect and input
 * commands using a command-line interface to the robot.
 *
 *                             ** COPYRIGHT NOTICE **
 *
 * Copyright (c) 2000, Nomadic Technologies, Inc.
 *
 * The contents of this file is copyrighted by Nomadic Technologies, Inc, and
 * is protected by United States copyright laws. You may make as many copies of
 * the software as deemed necessary for the effective programming of the Nomad
 * series robots provided that all copyright and other proprietary notices are
 * reproduced in their entirety.
 *
 */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>

#include "robotd.h"

#define bit(map, pos) ((map >> pos) & 1)
#define yn(map, pos) (bit(map, pos) ? 'y' : 'n')
#define batt(map) (bit(map, 3) ? "High" : (bit(map, 2) ? "Med" : "Low"))
#define le_to_short(buf, start)   ((buf[start+1] <= 0x7f) ?                    \
   ((unsigned short)buf[start] + (((unsigned short)buf[start+1]) << 8)) :      \
   (32768 - (unsigned short)buf[start] - (((unsigned short)buf[start+1]) << 8)))

unsigned char outbuf[BUFLEN];
extern int debug;


#if 0
static void print_buf(unsigned char *buf, int len)
{
  int ii;
  for (ii = 0; ii < len; ii++)
    printf("%02x ", buf[ii]);
  printf("\n");
}
#endif

static void debug_out(unsigned char *buf, int len)
{
#if 0
  if (!debug)
    return;
  printf("out: ");
  print_buf(buf, len);
#else
  return;
#endif
}


static void debug_in(unsigned char *buf, int len)
{
#if 0
  if (!debug)
    return;
  printf(" in: ");
  print_buf(buf, len);
#else
  return;
#endif
}


/* construct_hdr(): given an already-existing buffer, length and type,
 * this function initializes the header of the packet and sets the index to
 * immediately follow the header.  note that this is a byte-index.
 */
static void construct_hdr(struct msg_hdr *hdr, int len, int type, int *idx)
{
  hdr->magic = 1;
  hdr->type = type;
  hdr->len0 = len & 0xff;
  hdr->len1 = (len & 0xff00) >> 8;
  *idx = sizeof(struct msg_hdr);
}


/* append_byte(): given a buffer representing a robot comm packet and a byte
 * of data, append the byte to the packet and increment the index
 */
static void append_byte(unsigned char buf[BUFLEN], unsigned char byte, int *idx)
{
  buf[(*idx)++] = byte;
}


/* append_short(): given a buffer representing a robot comm packet and 2 bytes
 * of data, append the bytes to the packet and increment the index
 */
static void append_short(unsigned char buf[BUFLEN], short val, int *idx)
{
  if (val < 0)
    val = 32768 - val;
  append_byte(buf, val & 0xff, idx);
  append_byte(buf, (val >>= 8) & 0xff, idx);
}


/* append_int(): given a buffer representing a robot comm packet and 4 bytes
 * of data, append the bytes to the packet and increment the index
 */
#if 0
static void append_int(unsigned char buf[BUFLEN], unsigned int val, int *idx)
{
  append_byte(buf, val & 0xff, idx);
  append_byte(buf, (val >>= 8) & 0xff, idx);
  append_byte(buf, (val >>= 8) & 0xff, idx);
  append_byte(buf, (val >>= 8) & 0xff, idx);
}
#endif


/* append_checksum(): sum all previous bytes in buffer and append the lower
 * 8 bits of the sum to the buffer.
 */
static void append_checksum(unsigned char buf[BUFLEN], int *idx)
{
  int ii, sum = 0;
  for (ii = 0; ii < *idx; ii++)
    sum += buf[ii];
  append_byte(buf, sum & 0xff, idx);
}


/* append_end(): append the constant byte value 92 to the buffer. */
static void append_end(unsigned char buf[BUFLEN], int *idx)
{
  append_byte(buf, 92, idx);
}


static void finish_pkt(int fd, int *idx) 
{
  append_checksum(outbuf, idx);
  append_end(outbuf, idx);
  debug_out(outbuf, *idx);
  write(fd, outbuf, *idx);
  HOST_Forward(-1, fd);
}


static void finish_and_recv(int fd, int *idx) 
{
  append_checksum(outbuf, idx);
  append_end(outbuf, idx);
  debug_out(outbuf, *idx);
  write(fd, outbuf, *idx);
  HOST_ReadN(fd, outbuf, 4);
  HOST_ReadN(fd, outbuf + 4, outbuf[1]);
  debug_in(outbuf, outbuf[1]+4);
}


static int print(int fd, char *str) 
{
  return write(fd, str, strlen(str));
}


/* get_args(): use strtok() to get arguments from the command line and parse
 * them into integers.  if insufficient inputs are availalbe, returns -1, else
 * returns the number read.
 */
static int get_args(int *args, int cnt, char *name, int user)
{
  int ii;
  char *tok;
  for (ii = 0; ii < cnt; ii++) {
    tok = strtok(NULL, whitespace);
    if (!tok) {
      char out[128];
      sprintf(out, "\nInsufficient arguments.  %s requires %d args.\n\n", 
              name, cnt);
      print(user, out);
      return -1;
    }
    if (sscanf(tok, "%d", &args[ii]) != 1) {
      char out[128];
      sprintf(out, "\nArgument %d contains garbage.\n\n", ii + 1);
      print(user, out);
      return -1;
    }
  }
  return ii;
}


/* get_args_range(): use strtok() to get arguments from the command line and 
 * parse * them into integers.  if insufficient inputs are availalbe, 
 * returns -1, else returns the number read.
 */
static int get_args_range(int *args, int min, int max, char *name, int user)
{
  int ii;
  char *tok;
  for (ii = 0; ii < max; ii++) {
    tok = strtok(NULL, whitespace);
    if (!tok) {
      if (ii >= min)
        return ii;
      else {
        char out[128];
        sprintf(out, "\nInsufficient arguments.  %s requires %d args.\n\n", 
                name, min);
        print(user, out);
        return -1;
      }
    }
    if (sscanf(tok, "%d", &args[ii]) != 1) {
      char out[128];
      sprintf(out, "\nArgument %d contains garbage.\n\n", ii + 1);
      print(user, out);
      return -1;
    }
  }
  return ii;
}


static int check_args_min(int *args, int cnt, int min, int user)
{
  int idx;
  char out[128];
  for (idx = 0; idx < cnt; idx++)
    if (args[idx] < min) {
      sprintf(out, "\nArgument %d is out of range.  All args must be >= %d.\n\n", idx+1, min);
      print(user, out);
      return -1;
    }
  return 0;
}


static int check_args_max(int *args, int cnt, int max, int user)
{
  int idx;
  char out[128];
  for (idx = 0; idx < cnt; idx++)
    if (args[idx] > max) {
      sprintf(out, "\nArgument %d is out of range.  All args must be <= %d.\n\n", idx+1, max);
      print(user, out);
      return -1;
    }
  return 0;
}


static void do_AC(int fd, int user)
{
  int idx;
  int args[2];

  if (get_args(args, 2, "AC", user) < 0)
    return;

  if (check_args_min(args, 2, 1, user)
      || check_args_max(args, 2, 32767, user))
    return;

  construct_hdr((struct msg_hdr*)outbuf, 8, AC, &idx);
  append_short(outbuf, args[0], &idx);
  append_short(outbuf, args[1], &idx);
  append_short(outbuf, 0, &idx);
  finish_pkt(fd, &idx);
  print(user, "\nAcceleration set.\n\n");
}


static void do_DP(int fd, int user)
{
  int idx;
  int args[2];

  if (get_args(args, 2, "DP", user) < 0)
    return;

  if (check_args_min(args, 2, -32767, user)
      || check_args_max(args, 2, 32767, user))
    return;

  construct_hdr((struct msg_hdr*)outbuf, 6, DP, &idx);
  append_short(outbuf, args[0], &idx);
  append_short(outbuf, args[0], &idx);
  finish_pkt(fd, &idx);
  print(user, "\nPosition set.\n\n");
}


static void do_DA(int fd, int user)
{
  int idx;
  int args[1];

  if (get_args(args, 1, "DA", user) < 0)
    return;

  if (check_args_min(args, 1, 0, user)
      || check_args_max(args, 1, 3599, user))
    return;

  construct_hdr((struct msg_hdr*)outbuf, 6, DA, &idx);
  append_short(outbuf, args[0], &idx);
  append_short(outbuf, 0, &idx);
  finish_pkt(fd, &idx);
  print(user, "\nAngle set.\n\n");
}


/* do_GS(): send a get-state packet to the robot and display the reply.
 *          the format of the packet is as follows 
 *          (note: all values in little-endian order):
 * start	end	description
 * 0		3	header (0x01, length (2 bytes), type)
 * 4		19	sonar 
 * 20		22	bumper
 * 23		30	configuration (x, y, steer, "turret" -- 2 bytes each)
 * 31		36	current velocity (right, left, ignore -- 2 bytes each)
 * 37		38	compass -- ignore this
 * 39		39	status byte
 * 40		41	end-of-packet data (cksum, 0x5c
 */
static void do_GS(int fd, int user)
{
  int idx;
  char out[128];
  construct_hdr((struct msg_hdr*)outbuf, 2, GS, &idx);
  finish_and_recv(fd, &idx);
  print(user, "\n   Sonar data:");
  for (idx = 4; idx < 20; idx++) {
    sprintf(out, "%4d", outbuf[idx]);
    print(user, out);
  }
  print(user, "\n\n   Bumper Data : ");
  for (idx = 5; idx >= 0; idx--) {
    sprintf(out, "%d", (outbuf[20] >> idx) & 1);
    print(user, out);
  }
  sprintf(out, "\n\n   X Pos:%7d,  Steer Pos:%7d\n", le_to_short(outbuf, 23),
         le_to_short(outbuf, 27));
  print(user, out);
  sprintf(out, "   Y Pos:%7d\n   Right Wheel Speed:%4d, Left Wheel Speed:%4d\n",
          le_to_short(outbuf, 25), le_to_short(outbuf, 31), 
          le_to_short(outbuf, 33));
  print(user, out);
  sprintf(out, "\n   Motors: R:%c  L:%c   Battery:%s   Plugged-In:%c   Charging:%c   E-stop:%c",
          yn(outbuf[39], 0), yn(outbuf[39], 1), batt(outbuf[39]), 
          yn(outbuf[39], 4), yn(outbuf[39], 5), yn(outbuf[39], 6));
  print(user, out);
  print(user, "\n\n");
}


static void do_LP(int fd, int user)
{
  int idx;
  construct_hdr((struct msg_hdr*)outbuf, 2, LP, &idx);
  finish_pkt(fd, &idx);
  print(user, "\nEntering LIMP mode.\n\n");
}


static void do_PR(int fd, int user)
{
  int idx;
  int args[2];

  if (get_args(args, 2, "PR", user) < 0)
    return;

  if (check_args_min(args, 2, -32767, user)
      || check_args_max(args, 2, 32767, user))
    return;

  construct_hdr((struct msg_hdr*)outbuf, 8, PR, &idx);
  append_short(outbuf, args[0], &idx);
  append_short(outbuf, args[1], &idx);
  append_short(outbuf, 0, &idx);
  finish_pkt(fd, &idx);
  print(user, "\nMoving...\n\n");
}


static void do_SP(int fd, int user)
{
  int idx;
  int args[2];

  if (get_args(args, 2, "SP", user) < 0)
    return;

  if (check_args_min(args, 2, 1, user)
      || check_args_max(args, 2, 32767, user))
    return;

  construct_hdr((struct msg_hdr*)outbuf, 8, SP, &idx);
  append_short(outbuf, args[0], &idx);
  append_short(outbuf, args[1], &idx);
  append_short(outbuf, 0, &idx);
  finish_pkt(fd, &idx);
  print(user, "\nSpeed set.\n\n");
}


static void do_ST(int fd, int user)
{
  int idx;
  construct_hdr((struct msg_hdr*)outbuf, 2, ST, &idx);
  finish_pkt(fd, &idx);
  print(user, "\nStopping.\n\n");
}


static void do_VM(int fd, int user)
{
  int idx;
  int args[2];

  if (get_args(args, 2, "VM", user) < 0)
    return;

  if (check_args_min(args, 2, -32767, user)
      || check_args_max(args, 2, 32767, user))
    return;

  construct_hdr((struct msg_hdr*)outbuf, 8, VM, &idx);
  append_short(outbuf, args[0], &idx);
  append_short(outbuf, args[1], &idx);
  append_short(outbuf, 0, &idx);
  finish_pkt(fd, &idx);
  print(user, "\nMoving...\n\n");
}


static void do_WS(int fd, int user)
{
  int idx;
  int args[4];

  if ((get_args(args, 3, "WS", user)) < 0)
    return;

  construct_hdr((struct msg_hdr*)outbuf, 6, WS, &idx);
  append_byte(outbuf, args[0], &idx);
  append_byte(outbuf, args[1], &idx);
  append_byte(outbuf, 0, &idx);
  append_byte(outbuf, args[2], &idx);
  finish_and_recv(fd, &idx);
  if ((args[0] && bit(outbuf[39], 0)) || (args[1] && bit(outbuf[39], 1)))
    print(user, "\nNot stopped.\n\n");
  else
    print(user, "\nStopped.\n\n");
}


static void do_ZR(int fd, int user)
{
  int idx;
  construct_hdr((struct msg_hdr*)outbuf, 2, ZR, &idx);
  finish_pkt(fd, &idx);
  print(user, "\nZeroed.\n\n");
}


static void do_CONF_SN(int fd, int user)
{
  int idx, ii;
  int args[17] = {0,}, arg;

  if ((arg = get_args_range(args, 2, 17, "CONF_SN", user)) < 0)
    return;

  construct_hdr((struct msg_hdr*)outbuf, 19, CONF_SN, &idx);
  if (arg < 17 && args[arg-1] >= 0) {
    print(user, "\nInsufficient arguments.  CONF_SN requires at least 2 args and the last\n  one must be negative if there are fewer than 17.\n\n");
    return;
  }
  for (ii = 0; ii < 17; ii++) {
    if (ii > 0 && args[ii] >= 16) {
      char out[128];
      sprintf(out, "\nArgument %d is out of range.\n\n", ii+1);
      print(user, out);
      return;
    }
    if (args[ii] < 0) /* negative (terminator) is mapped to 0xff */
      append_byte(outbuf, (unsigned char)0xff, &idx);
    else
      append_byte(outbuf, (unsigned char)args[ii], &idx);
  }
  finish_pkt(fd, &idx);
  print(user, "\nSonar configured.\n\n");
}


static void do_CONF_TM(int fd, int user)
{
  int idx;
  int args[1];

  if (get_args(args, 1, "CONF_TM", user) < 0)
    return;

  if (check_args_min(args, 1, 0, user)
      || check_args_max(args, 1, 255, user))
    return;

  construct_hdr((struct msg_hdr*)outbuf, 3, CONF_TM, &idx);
  append_byte(outbuf, args[0], &idx);
  finish_pkt(fd, &idx);
  print(user, "\nTimeout configured.\n\n");
}


static void do_GET_RA(int fd, int user)
{
  int idx;
  char out[128];
  construct_hdr((struct msg_hdr*)outbuf, 2, GET_RA, &idx);
  finish_and_recv(fd, &idx);
  sprintf(out, "\nRight Acceleration: %d\n", le_to_short(outbuf, 4));
  print(user, out);
  sprintf(out, " Left Acceleration: %d\n\n", le_to_short(outbuf, 6));
  print(user, out);
}


static void do_GET_RC(int fd, int user)
{
  int idx;
  char out[128];
  construct_hdr((struct msg_hdr*)outbuf, 2, GET_RC, &idx);
  finish_and_recv(fd, &idx);
  sprintf(out, "\n    X Position: %d\n", le_to_short(outbuf, 4));
  print(user, out);
  sprintf(out, "    Y Position: %d\n", le_to_short(outbuf, 6));
  print(user, out);
  sprintf(out, "Steer Position: %d\n\n", le_to_short(outbuf, 8));
  print(user, out);
}


static void do_GET_RV(int fd, int user)
{
  int idx;
  char out[128];
  construct_hdr((struct msg_hdr*)outbuf, 2, GET_RV, &idx);
  finish_and_recv(fd, &idx);
  sprintf(out, "\n\nRight Wheel Speed: %d\n", le_to_short(outbuf, 4));
  print(user, out);
  sprintf(out, " Left Wheel Speed: %d\n\n", le_to_short(outbuf, 6));
  print(user, out);
}


static void do_GET_SN(int fd, int user)
{
  int idx;
  char out[128];
  construct_hdr((struct msg_hdr*)outbuf, 2, GET_SN, &idx);
  finish_and_recv(fd, &idx);
  print(user, "\n   Sonar data:");
  for(idx = 4; idx < 20; idx++) {
    sprintf(out, "%4d", outbuf[idx]);
    print(user, out);
  }
  print(user, "\n\n");
}


static void do_INFO(int user)
{
  print(user, "Robot Command Interpreter, Version 1.0\n");
  print(user, "(C) Copyright 1996-1999, Nomadic Technologies, Inc.\n");
  print(user, "Phone: 650/988-7200  Fax: 650/988-7201\n");
  print(user, "Email: nomad@robots.com\n");
  print(user, "\nYou are connected on the console on port 65000.\n\n");
}


/* do_HELP(): print help for requested command. */
static void do_HELP(int user)
{
  char *tok = strtok(NULL, whitespace);

  if (tok == NULL || !strncmp(tok, "HELP", 4) || !strncmp(tok, "INFO", 4)) {
    print(user, "\nAvailable commands:\n");
    print(user, "     AC      DP      DA      GS      LP      PR      SP      ST\n");
    print(user, "     VM      WS      ZR      CONF_SN CONF_TM GET_RA  GET_RC  GET_RV\n");
    print(user, "     GET_SN  INFO    HELP\n");
    print(user, "\nType the command name after HELP to get more detailed information.\n\n");
  } else if (!strncmp(tok, "AC", 2)) {
    print(user, "\nAC\n     This sets the right and left wheel accelerations of the\n     robot to the values given.\n\n");
  } else if (!strncmp(tok, "DP", 2)) {
    print(user, "\nDP\n     This will set the robot encoder position to the given X and Y values.\n\n");
  } else if (!strncmp(tok, "DA", 2)) {
    print(user, "\nDA\n     This will set the robot encoder rotation to the given steering value.\n\n");
  } else if (!strncmp(tok, "GS", 2)) {
    print(user, "\nGS\n     This will display the robot's current state information.\n\n");
  } else if (!strncmp(tok, "LP", 2)) {
    print(user, "\nLP\n     This will set the motors to limp mode (turn them off).\n\n");
  } else if (!strncmp(tok, "PR", 2)) {
    print(user, "\nPR\n     This will move the robot to a relative encoder position.  Arguments\n     are desired right and left wheel relative movements.\n\n");
  } else if (!strncmp(tok, "SP", 2)) {
    print(user, "\nSP\n     This sets the right and left wheel speeds of the motors.\n\n");
  } else if (!strncmp(tok, "ST", 2)) {
    print(user, "\nST\n     This command will bring the robot to a stop, but leave the motors on.\n\n");
  } else if (!strncmp(tok, "VM", 2)) {
    print(user, "\nVM\n     This will start the motors moving in an absolute velocity mode.\n     Arguments are right and left wheel velocities, in units\n     of 0.1 inches or degrees per second.\n\n");
  } else if (!strncmp(tok, "WS", 2)) {
    print(user, "\nWS\n     This will wait for the given timeout period or until the specified\n     motors are stopped, whichever comes first.\n\n");
  } else if (!strncmp(tok, "ZR", 2)) {
    print(user, "\nZR\n     This will reset the encoder values to zero.\n\n");
  } else if (!strncmp(tok, "CONF_SN", 7)) {
    print(user, "\nCONF_SN\n     This sets the sonar firing order.  First value is firing rate,\n     others are firing values.  If you use less than 16, give -1 to terminate.\n\n");
  } else if (!strncmp(tok, "CONF_TM", 7)) {
    print(user, "\nCONF_TM\n     This sets the timeout value in seconds.  A value of zero disables timeout.\n\n");
  } else if (!strncmp(tok, "GET_RA", 6)) {
    print(user, "\nGET_RA\n     Display robot acceleration data.\n\n");
  } else if (!strncmp(tok, "GET_RC", 6)) {
    print(user, "\nGET_RC\n     Display robot configuration information.\n\n");
  } else if (!strncmp(tok, "GET_RV", 6)) {
    print(user, "\nGET_RV\n     Display motor velocities.\n\n");
  } else if (!strncmp(tok, "GET_SN", 6)) {
    print(user, "\nGET_SN\n     Display sonar values.\n\n");
  }
  else {
    char out[128];
    sprintf(out, "\nI don't understand \"%s\".\n\n", tok);
    print(user, out);
  }
}


/* process_console_msg(): read command lines from the console and execute
 * the appropriate command.  this function assumes that the incoming telnet
 * is in line-oriented mode.  otherwise editing of text will not be possible.
 */
int CON_ProcessMsg(int user, int fd)
{
  unsigned char buf[BUFLEN];
  char *pos;    /* position of \n char */
  char *tok;	/* current string token */
  int len, tot = 0, ii;

  /* read input until we have a whole line of text */
  do {
    if ((len = read(user, buf + tot, BUFLEN)) <= 0)
      return -1;
    tot += len;
  } while ((pos = strchr(buf, '\n')) == NULL);

  /* mark \n as the end of the input by turning it into a NULL char */
  *pos = 0;

  /* uppercase the whole thing */
  for (ii = 0; buf[ii]; ii++)
    buf[ii] = toupper(buf[ii]);

  /* begin tokenizing input -- first word is command */
  tok = strtok(buf, whitespace);
#if 0
  if (debug) printf("console cmd is %s\n", tok);
#endif
  if (tok) {
    if (!strncmp(tok, "AC", 2)) {
      do_AC(fd, user);
    } else if (!strncmp(tok, "DP", 2)) {
      do_DP(fd, user);
      } else if (!strncmp(tok, "DA", 2)) {
      do_DA(fd, user);
    } else if (!strncmp(tok, "GS", 2)) {
      do_GS(fd, user);
    } else if (!strncmp(tok, "LP", 2)) {
      do_LP(fd, user);
    } else if (!strncmp(tok, "PR", 2)) {
      do_PR(fd, user);
    } else if (!strncmp(tok, "SP", 2)) {
      do_SP(fd, user);
    } else if (!strncmp(tok, "ST", 2)) {
      do_ST(fd, user);
    } else if (!strncmp(tok, "VM", 2)) {
      do_VM(fd, user);
    } else if (!strncmp(tok, "WS", 2)) {
      do_WS(fd, user);
    } else if (!strncmp(tok, "ZR", 2)) {
      do_ZR(fd, user);
    } else if (!strncmp(tok, "CONF_SN", 7)) {
      do_CONF_SN(fd, user);
    } else if (!strncmp(tok, "CONF_TM", 7)) {
      do_CONF_TM(fd, user);
    } else if (!strncmp(tok, "GET_RA", 6)) {
      do_GET_RA(fd, user);
    } else if (!strncmp(tok, "GET_RC", 6)) {
      do_GET_RC(fd, user);
    } else if (!strncmp(tok, "GET_RV", 6)) {
      do_GET_RV(fd, user);
    } else if (!strncmp(tok, "GET_SN", 6)) {
      do_GET_SN(fd, user);
    } else if (!strncmp(tok, "INFO", 4)) {
      do_INFO(user);
    } else if (!strncmp(tok, "HELP", 4)) {
      do_HELP(user);
    } else if (!strcmp(tok, "QUIT") || !strcmp(tok, "EXIT")) {
      print(user, "Goodbye\n");
      return -1;	/* close the connection */
    } else {
      char out[128];
      sprintf(out, "I don't understand \"%s\".\n", tok);
      print(user, out);
    }
  }
  print(user, "Robot-> ");
  return 1;
}


void CON_PrintGreeting(int user)
{
  print(user, "\nScout version 1.1 initializing...\n");
  print(user, "* * * * * * * * * * *\n\n");
  print(user, "Scout Robot Command Interpreter, Version 1.0\n\n");
  print(user, "(C) Copyright 1999, Nomadic Technologies, Inc.\n\n");
  print(user, "Type HELP for help.\n");
  print(user, "Robot-> ");
}
