/*
 * scout-robotd/main.c
 *
 *                             ** COPYRIGHT NOTICE **
 *
 * Copyright (c) 2000, Nomadic Technologies, Inc.
 *
 * The contents of this file is copyrighted by Nomadic Technologies, Inc, and
 * is protected by United States copyright laws. You may make as many copies of
 * the software as deemed necessary for the effective programming of the Nomad
 * series robots provided that all copyright and other proprietary notices are
 * reproduced in their entirety.
 *
 */

#include <errno.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <ctype.h>
#include <termios.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <signal.h>

#include "robotd.h"
#include "setup.h"

#define max(aa, bb)  ((aa) > (bb) ? (aa) : (bb))

#define ROBD_DEFAULT_ROBOT_PORT        4001
#define ROBD_DEFAULT_CONSOLE_PORT      4000
#define ROBD_DEFAULT_SERIAL_FILENAME   "/dev/ttyS0"
#define ROBD_DEFAULT_SETUP_FILENAME    "/etc/robotd.cfg"



/* command-line arguments (some may be overriden by setup file) */
static const char *robd_setup_file = NULL,
  *robd_robot_portname = NULL, *robd_con_portname = NULL,
  *robd_serial_filename = NULL;
static int robd_help_needed = 0, robd_no_deamon = 0;

/* all the above command-line options override setup file options, excepting
 * the robd_help_needed, robd_no_deamon, and robd_setup_file which
 * have no setup file entries.  the remaining entries in the setup file
 * (with no command-line equivilants) are below: */
static const char *robd_speech_command;

static int robd_done;
static long robd_host_port, robd_con_port;




/* ------------------------------------------------------------------------
 * Function:     Usage
 * Purpose:      Display the command-line options and exit
 * ------------------------------------------------------------------------ */
static void Usage(int exitcode)
{
  FILE *fp;

  fp = (exitcode == 0) ? stdout : stderr;
  fprintf(fp, "Usage: %s [options]\nOptions:\n"
	  "--help                display this message\n"
	  "--D                   do not fork to run in the background\n"
	  "--setupfile name      specify an alternate setup file location\n"
	  "--robotport port      specify the port to listen on for "
	  "client connections\n" 
	  "--consoleport port    specify the port to listen on for "
	  "console connections\n"
	  "--hostport filename   specify the serial port device for "
	  "host connection\n",
	  robd_program_name);
  exit(exitcode);
}



/* ------------------------------------------------------------------------
 * Function:     ParseArgs
 * Purpose:      Process the command line for the server
 * Arguments:    int Argc : the number of tokens
 *               char **Argv : the command line
 * Return:       int : 0, or -1 if bad parse
 * ------------------------------------------------------------------------ */
static int ParseArgs(int argc, char **argv)
{
  int i;
  char *arg;

  for (i = 1; i < argc; ++i)
  {
    arg = argv[i];

    /* check for flags (-option or --option) */
    if (*arg == '-')
    {
      if (*++arg == '-')
      {
	++arg;
      }

      /* check for setup file */
      if (strcasecmp(arg, "file") == 0 ||
	  strcasecmp(arg, "setupfile") == 0)
      {
	if (++i < argc)
	{
	  robd_setup_file = argv[i];
	}
	else
	{
	  return -1;
	}
      }
      else if (*arg == 'f')
      {
	if (strlen(arg) > 1)
	{
	  robd_setup_file = arg + 1;
	}
	else if (++i < argc)
	{
	  robd_setup_file = argv[i];
	}
	else
	{
	  return -1;
	}
      }

      /* check for local port */
      else if (strcasecmp(arg, "robotport") == 0)
      {
	if (++i < argc)
	{
	  robd_robot_portname = argv[i];
	}
	else
	{
	  return -1;
	}
      }
      else if (*arg == 'r')
      {
	if (strlen(arg) > 1)
	{
	  robd_robot_portname = arg + 1;
	}
	else if (++i < argc)
	{
	  robd_robot_portname = argv[i];
	}
	else
	{
	  return -1;
	}
      }

      else if (strcasecmp(arg, "consoleport") == 0)
      {
	if (++i < argc)
	{
	  robd_con_portname = argv[i];
	}
	else
	{
	  return -1;
	}
      }
      else if (*arg == 's')
      {
	if (strlen(arg) > 1)
	{
	  robd_con_portname = arg + 1;
	}
	else if (++i < argc)
	{
	  robd_con_portname = argv[i];
	  {
	    return -1;
	  }
	}
	else
	{
	  return -1;
	}
      }

      else if (strcasecmp(arg, "hostport") == 0 ||
	       strcasecmp(arg, "serialport") == 0)
      {
	if (++i < argc)
	{
	  robd_serial_filename = argv[i];
	}
	else
	{
	  return -1;
	}
      }
      else if (*arg == 's')
      {
	if (strlen(arg) > 1)
	{
	  robd_serial_filename = arg + 1;
	}
	else if (++i < argc)
	{
	  robd_serial_filename = argv[i];
	}
	else
	{
	  return -1;
	}
      }

      /* check for help request */
      else if (strcasecmp(arg, "help") == 0 ||
	       strcasecmp(arg, "?") == 0)
      {
	robd_help_needed = 1;
      }

      else if (strcasecmp(arg, "fg") == 0 ||
	       strcasecmp(arg, "d") == 0)
      {
	robd_no_deamon = 1;
      }
    }

    /* we currenly don't support non-flags */
    else
    {
      return -1;
    }
  }

  return 0;
}



static int ConvertArgs(void)
{
  char *tmp_string;

  if (robd_robot_portname != NULL)
  {
    /* check command line scheduler port number */
    robd_host_port = strtol(robd_robot_portname, &tmp_string, 0);
    if (*tmp_string != '\0' || robd_host_port < 0 || robd_host_port > 65535)
    {
      fprintf(stderr, "%s: \"%s\" is an invalid scheduler port number.  It "
	      "must be a valid integer in\n  the range [0 - 65535].\n",
	      robd_program_name, robd_robot_portname);
      return -1;
    }
  }
  else if ((robd_robot_portname = SETUP_GetValue("[connect]serv_port")) != NULL)
  {
    robd_host_port = strtol(robd_robot_portname, &tmp_string, 0);
    if (*tmp_string != '\0' || robd_host_port < 0 || robd_host_port > 65535)
    {
      fprintf(stderr, "%s: Invalid entry \"%s\" in setup file:\n  "
	      "[connect]serv_port must be a valid integer in the range "
	      "[0, 65535].\n", robd_program_name, robd_robot_portname);
      return -1;
    }
  }
  else
  {
    robd_host_port = ROBD_DEFAULT_ROBOT_PORT;
  }

  if (robd_con_portname != NULL)
  {
    /* check command line scheduler port number */
    robd_con_port = strtol(robd_con_portname, &tmp_string, 0);
    if (*tmp_string != '\0' || robd_con_port < 0 || robd_con_port > 65535)
    {
      fprintf(stderr, "%s: \"%s\" is an invalid scheduler port number.  It "
	      "must be a valid integer in\n  the range [0 - 65535].\n",
	      robd_program_name, robd_con_portname);
      return -1;
    }
  }
  else if ((robd_con_portname =
	    SETUP_GetValue("[connect]console_port")) != NULL)
  {
    robd_con_port = strtol(robd_con_portname, &tmp_string, 0);
    if (*tmp_string != '\0' || robd_con_port < 0 || robd_con_port > 65535)
    {
      fprintf(stderr, "%s: Invalid entry \"%s\" in setup file:\n  "
	      "[connect]console_port must be a valid integer in the range "
	      "[0, 65535].\n", robd_program_name, robd_con_portname);
      return -1;
    }
  }
  else
  {
    robd_con_port = ROBD_DEFAULT_CONSOLE_PORT;
  }

  if (robd_serial_filename == NULL)
  {
    if ((robd_serial_filename = SETUP_GetValue("[connect]serial_port")) == NULL)
    {
      robd_serial_filename = ROBD_DEFAULT_SERIAL_FILENAME;
    }
  }

  robd_speech_command = SETUP_GetValue("[speech]command");

  return 0;
}



static void ParseTag(char *tag, char **ver, char **note)
{
  char *cp;
  static char tagname[12];

  /* we parse the CVS tag for display.  If there is no Name tag
   * we display "UNSET"; if there is nothing but whitespace inside the
   * frame, we display "UNRELEASED"; otherwise, the first token
   * (delimited by whitespace or '=') is taken to be the version and
   * all '_' characters are replaced with '.'; the remaining tokens
   * are extra and all '_' characters are converted to ' '. */
  if (strlen(tag) < 7 || memcmp(tag, "$Name: ", 7) != 0 ||
      (cp = strchr((char *)tag + 7, '$')) == NULL)
  {
    strcpy(tagname, "UNSET");
    *ver = tagname;
    *note = NULL;
  }
  else
  {
    *cp = '\0'; /* null terminate, removing the ending '$' */
    cp = tag + 7; /* start after "$Name: " */
    while (isspace(*cp))
    {
      ++cp; /* remove leading whitespace */
    }
    if (*cp == '\0')
    {
      /* if all whitespace in the name string, there is no tag */
      strcpy(tagname, "UNRELEASED");
      *ver = tagname;
      *note = NULL;
    }
    else
    {
      *ver = *note = cp;
      while (**note != '\0')
      {
        if (isspace(**note) || **note == '=')
        {
          **note++ = '\0';
          break;
        }
        if (**note == '_')
        {
          **note = '.';
        }
        ++*note;
      }
      while (isspace(**note))
      {
	++note;
      }
      if (**note == '\0')
      {
        *note = NULL;
      }
      else
      {
        cp = *note;
        while (*cp != '\0')
        {
          if (*cp == '_')
          {
            *cp = ' ';
          }
          ++cp;
        }
      }
    }
  }

  return;
}



static void SignalTrap(int signal_num)
{
  if (robd_done == 0)
  {
    fprintf(stderr, "%s: exiting on signal %d.\n", robd_program_name,
	    signal_num);
    robd_done = 1;
  }
  return;
}



static int OpenSerial(const char *port)
{
  struct termios info;
  int fd;

  if ((fd = open(port, O_RDWR|O_NONBLOCK)) < 0)
  {
    fprintf(stderr, "%s: error opening %s: %s\n", robd_program_name, port,
	    strerror(errno));
    return -1;
  }

  if (tcgetattr(fd, &info) < 0)
  {
    fprintf(stderr, "%s: error in tcgetattr(): %s\n", robd_program_name,
	    strerror(errno));
    close(fd);
    return -1;
  }
 
  /* restore old values to unhang, if hung */
  info.c_iflag=1280;
  info.c_oflag=5;
  info.c_cflag=3261;
  info.c_lflag=35387;

  if (tcsetattr(fd, TCSANOW, &info) < 0)
  {
    fprintf(stderr, "%s: error in tcsetattr(): %s\n", robd_program_name,
	    strerror(errno));
    close(fd);
    return -1;
  }

  info.c_iflag = 0;
  info.c_oflag = 0;
  info.c_lflag = 0;
  info.c_cflag = B38400 | CS8 | CREAD | CLOCAL;

  /* set time out on serial read */
  info.c_cc[VMIN] = 0;
  info.c_cc[VTIME] = 10;

  if (tcsetattr(fd,TCSANOW,&info) < 0)
  {
    fprintf(stderr, "%s: error in tcsetattr(): %s\n", robd_program_name,
	    strerror(errno));
    close(fd);
    return -1;
  }

  return fd;
}



/* open_socket(): open a TCP socket on a given port */
static int OpenListenerSocket(int port)
{
  struct sockaddr_in server;
  int fd, i;

  fd = socket(AF_INET, SOCK_STREAM, 0);
  if (fd < 0)
  {
    fprintf(stderr, "%s: error allocating listener socket for port %d: %s\n",
	    robd_program_name, port, strerror(errno));
    return -1;
  }

  i = 1;
  i = setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (void *)&i, sizeof(i));
  if (i < 0)
  {
    fprintf(stderr, "%s: setsockopt() error for port %d: %s\n",
	    robd_program_name, port, strerror(errno));
    close(fd);
    return -1;
  }

  server.sin_family = AF_INET;
  server.sin_port = htons(port);
  server.sin_addr.s_addr = INADDR_ANY;
  if (bind(fd, &server, sizeof(server)) < 0)
  {
    fprintf(stderr, "%s: error binding port %d: %s\n", robd_program_name,
	    port, strerror(errno));
    close(fd);
    return -1;
  }

  listen(fd, 4);
  return fd;
}





int main(int argc, char **argv)
{
  char *tag, *note, *cp, tag_buf[80];
  int lhostfd, lconfd, hostfd, confd, ttyfd, pid, maxfd, len, ret = 0;
  fd_set set;
  struct sockaddr_in client;

  cp = argv[0] + strlen(argv[0]) - 1;
  while (cp != argv[0])
  {
    if (*cp == '/')
    {
      ++cp;
      break;
    }
    --cp;
  }
  robd_program_name = cp;

  printf("%s: initializing...\n", robd_program_name);

  if (ParseArgs(argc, argv) < 0)
  {
    Usage(-1);
  }
  if (robd_help_needed)
  {
    Usage(0);
  }

  /* read in the user's setup file */
  if (robd_setup_file== NULL)
  {
    SETUP_ReadSetup(ROBD_DEFAULT_SETUP_FILENAME);
  }
  else
  {
    if (SETUP_ReadSetup(robd_setup_file) != 0)
    {
      fprintf(stderr, "%s (fatal): could not read %s.\n", robd_program_name,
	      robd_setup_file);
      exit(-1);
    }
  }
  
  if (ConvertArgs() != 0)
  {
    exit(-1);
  }

  /* open all ports */
  if ((ttyfd = OpenSerial(robd_serial_filename)) < 0)
  {
    exit(-1);
  }
  lhostfd = OpenListenerSocket((unsigned short)robd_host_port);
  lconfd = OpenListenerSocket((unsigned short)robd_con_port);
  if (hostfd < 0 && confd < 0)
  {
    exit(-1);
  }

  hostfd = -1;
  confd = -1;

  TK_Init(robd_speech_command);
    
  tag = "$Name:  $";
  if (strlen(tag) > sizeof(tag_buf) - 1)
  {
    strcpy(tag_buf, "REVISION TAG SIZE OVERFLOW");
    tag = tag_buf;
    note = NULL;
  }
  else
  {
    strcpy(tag_buf, tag);
    ParseTag(tag_buf, &tag, &note);
  }

  if (robd_no_deamon)
  {
    pid = -1;
  }
  else if ((pid = fork()) == -1)
  {
    fprintf(stderr, "%s: (warning) fork() failed: ", robd_program_name);
    perror(NULL);
  }

  if (pid != 0) /* parent process or error forking */
  {
    if (note != NULL)
    {
      printf("%s: version       %s (%s, %s)\n", robd_program_name,
	     tag, note, __DATE__);
    }
    else
    {
      printf("%s: version       %s (%s)\n", robd_program_name, tag, __DATE__);
    }
    printf("%s: robot port    %ld\n", robd_program_name, robd_host_port);
    printf("%s: console port  %ld\n", robd_program_name, robd_con_port);
    printf("%s: host port     %s\n", robd_program_name, robd_serial_filename);
    if (pid != -1)
    {
      printf("%s: process ID    %d\n", robd_program_name, pid);
    }
    printf("%s: initialized.\n\n", robd_program_name);

    if (pid != -1)
    {
      return 0;
    }
  }

  /* install signal handler for unexpected quits */
  robd_done = 0;
  signal(SIGINT, SignalTrap);
  signal(SIGTERM, SignalTrap);
  signal(SIGHUP, SignalTrap);

  /* loop until killed */
  while (robd_done == 0)
  {
    FD_ZERO(&set);
    maxfd = 0;
    if (hostfd != -1)
    {
      FD_SET(hostfd, &set);
      maxfd = max(maxfd, hostfd);
    }
    else if (lhostfd != -1)
    {
      FD_SET(lhostfd, &set);
      maxfd = max(maxfd, lhostfd);
    }
    if (confd != -1)
    {
      FD_SET(confd, &set);
      maxfd = max(maxfd, confd);
    }
    else if (lconfd != -1)
    {
      FD_SET(lconfd, &set);
      maxfd = max(maxfd, lconfd);
    }
    if ((len = select(maxfd + 1, &set, NULL, NULL, NULL)) < 0)
    {
      if (errno != EINTR)
      {
	fprintf(stderr, "%s: error in select(): %s\n", robd_program_name,
		strerror(errno));
	ret = -1;
      }
      break;
    }

    /* if there is a message on the server socket, forward it to the
       robot and wait for a reply */
    if (hostfd != -1 && FD_ISSET(hostfd, &set))
    {
      if (HOST_Forward(ttyfd, hostfd) <= 0 ||
	  HOST_Forward(hostfd, ttyfd) <= 0)
      {
        /* remote host closed connection */
        close(hostfd);
        hostfd = -1;
      }
    }

    /* if there's a message waiting on the console, process it */
    if (confd != -1 && FD_ISSET(confd, &set))
    {
      if (CON_ProcessMsg(confd, ttyfd) < 0)
      {
        /* remote host closed connection */
        close(confd);
        confd = -1;
      }
    }

    /* if there's a pending server connection, accept it */
    if (hostfd == -1 && lhostfd != -1 && FD_ISSET(lhostfd, &set))
    {
      if ((hostfd = accept(lhostfd, (struct sockaddr *)&client, &len)) < 0)
      {
	fprintf(stderr, "%s: error in accept(): %s\n", robd_program_name,
		strerror(errno));
	hostfd = -1;
      } 
    }

    /* if there's a pending console connection, accept it */
    if (confd < 0 && lconfd != -1 && FD_ISSET(lconfd, &set))
    {
      if ((confd = accept(lconfd, &client, &len)) < 0)
      {
	fprintf(stderr, "%s: error in accept(): %s\n", robd_program_name,
		strerror(errno));
      }
      else
      {
	CON_PrintGreeting(confd);
      }
    }
  }

  close(confd);
  close(ttyfd);
  close(hostfd);

  return ret;
}
