/*
 * scout-robotd/setup.c
 *
 * Cognos setup files reading and storage - dynamic categories and item
 * C source code
 *
 *                   ** COPYRIGHT NOTICE **
 *
 * Copyright (c) 1996, Nomadic Technologies, Inc.
 *
 * The contents of this file is copyrighted by Nomadic Technologies, Inc, and
 * is protected by United States copyright laws. You may make as many copies of
 * the software as deemed necessary for the effective programming of the Nomad
 * series robots provided that all copyright and other proprietary notices are
 * reproduced in their entirety.
 *
 */

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <assert.h>
#include <math.h>

#include "hash.h"
#include "setup.h"


#define LINE_MAX_LENGTH 200 /* Maximum length of a line in a setup file */

#define TABLE_SIZE 512 /* A power of two, the total number of items in a file */

#define ILLEGAL_CHARS "=\\[]=;\t\n" 


/* Special chars */
#define SECTION_LEFT_DELIMITER "["
#define SECTION_RIGHT_DELIMITER "]"
#define EQUAL "="
#define WHITE_SPACE " \n\t"
#define COMMENT ";"
#define CONTINUATION "\\"



static cache_ptr Table = NULL;



void Cleanup(char *String);



/***************
 * FUNCTION:     SETUP_ReadSetup
 * PURPOSE:      Read a setup file organized in sections and item and stores it 
 *               into a hash table.
 * ARGUMENTS:    char *filename : Setup file name
 * ALGORITHM:    Builds the key by concatenating the section name and the item
 *               name.
 * RETURN:       BOOL: 0 if opening the file was successful
 * SIDE EFFECT:  None
 * CALLS:        Cleanup
 * CALLED BY:    API
 ***************/
int SETUP_ReadSetup(const char *filename)
{
  char buf1[LINE_MAX_LENGTH], buf2[LINE_MAX_LENGTH], buf3[LINE_MAX_LENGTH],
    *buf;
  FILE *fp;
  char *Token, *Key, *Value, *Aux, *SectionName = buf3;
  long Length, KeyLength, SectionNameLength, TokenLength, l;
  
  if ((fp = fopen(filename, "r")) == NULL)
  {
    printf("Unable to read setup file %s\n", filename);
    return -1;
  }
  
  /* Create hash table: keys = section names concatenated with item names. */
  if (!Table)
  {
    /* This is the first time this function is called - create Table. */
    Table = hash_new(TABLE_SIZE, (hash_func_type)hash_string, 
		     (compare_func_type)compare_strings);
    assert(Table);
  }
  
  while (fgets(buf1, LINE_MAX_LENGTH-1, fp) != NULL)
  {
    buf = buf1;
    buf += strspn(buf, WHITE_SPACE); /* Skip leading whitespaces and tabs. */
    if (*buf == *COMMENT)
      continue;
    else if(strchr(buf, *SECTION_LEFT_DELIMITER))
    {
      /* Start a new section. */
      Token = strtok(buf, ILLEGAL_CHARS);
      if (!Token)
      {
	printf("Incorrect section name format for section %s\n", buf);
	SectionName = NULL; /* will discard everything until new section */
      }
      else
      {
	/* Build section name: a string [<section name>] */
	Cleanup(Token);
	Length = strlen(Token);
	*SectionName = *SECTION_LEFT_DELIMITER;
	memcpy(SectionName+1, Token, Length+1);
	*(SectionName+Length+1) = *SECTION_RIGHT_DELIMITER;
	*(SectionName+Length+2) = 0;
	SectionNameLength = Length+2;
      }
    }
    else if (strchr(buf, *EQUAL))
    {
      /* Found an item and have a valid section under examination. */
      
      /* Build the value: the string after EQUAL, verbatim (illegal chars
	 removed though). */
      /* Do the value before the key because Token building uses strtok which
	 wrecks buf. */
      Aux = strstr(buf, EQUAL)+1;
      Aux += strspn(Aux, WHITE_SPACE); /* Skip leading blanks and tabs */
      Length = strcspn(Aux, COMMENT);
      Value = (char *)malloc(Length+1);
      memcpy(Value, Aux, Length);
      *(Value+Length) = 0;
      while (strchr(Value, *CONTINUATION))
      {
	/* Skip the continuation character. */
	Length = strcspn(Value, CONTINUATION);
	/* Read the continuation. */
	if (fgets(buf2, LINE_MAX_LENGTH-1, fp) != NULL)
	{
	  l = strlen(buf2);
	  Value = (char *)realloc(Value, Length+l);
	  memcpy(Value+Length, buf2, l);
	  Length += l;
	}
	*(Value+Length) = 0;
      }
      Cleanup(Value); /* Remove illegal chars, especially tabs and newlines. */
      /* At some point, it would be better to replace tabs by spaces, and
	 multiple spaces by one. */
      
      /* Build the Key: a string [<section name>]<item name>. */
      Token = strtok(buf, ILLEGAL_CHARS);
      Cleanup(Token);
      TokenLength = strlen(Token);
      Key = (char *)malloc(SectionNameLength+TokenLength+1);
      memcpy(Key, SectionName, SectionNameLength);
      KeyLength = SectionNameLength;
      memcpy(Key+KeyLength, Token, strlen(Token));
      KeyLength += strlen(Token);
      *(Key+KeyLength) = 0; /* A good 'ol null-terminated string */


      /* Record the Value for the Key. */
      /* printf("Recording %s ---> %s\n", Key, Value);  */
      if (hash_value_for_key(Table, Key))
      {
	hash_change_key_value(Table, (void *)Key, (void *)Value);
	free(Key);
      }
      else
      {
	hash_add(&Table, (void *)Key, (void *)Value);
      }
    }
  }
  fclose(fp);

  return 0;
}

/***************
 * FUNCTION:     SETUP_GetValue
 * PURPOSE:      Get the value for a given key
 * ARGUMENTS:    char *Key : the key of the form [<section>]<item>
 * ALGORITHM:    Hash table call
 * RETURN:       const char *: the registered string DO NOT FREE IT !!!
 * SIDE EFFECT:  None
 * CALLS:        hash_value_for_key (common  package)
 * CALLED BY:    API
 ***************/
char *SETUP_GetValue(char *Key)
{
  if (!Table)
    return NULL;
  return (char *)hash_value_for_key(Table, Key);
}


char *SETUP_ExtGetValue(char *Section, char *Key)
{
  char *tmp, *ret;

  tmp = (char *)malloc(strlen(Section) + strlen(Key) + 3);
  sprintf(tmp, "[%s]%s", Section, Key);
  
  ret = SETUP_GetValue(tmp);
  free(tmp);
  return ret;
}



/***************
 * FUNCTION:     SETUP_SetValue
 * PURPOSE:      
 * ARGUMENTS:    char *Key : the key of tht form [<section>]<item>
 *               char *UserValue : the value the key will reference
 * ALGORITHM:    Hash table value change.
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/
void SETUP_SetValue(char *Key, char *UserValue)
{
  char *Value, *OldValue, *NewKey;
  
  assert(Key);
  assert(UserValue);
  
  Value = (char *)malloc(strlen(UserValue)+1);
  
  if (!Table)
  {
    /* This is the first time this function is called - create Table. */
    Table = hash_new(TABLE_SIZE, (hash_func_type)hash_string, 
		     (compare_func_type)compare_strings);
    assert(Table);
  }
  
  /* Create a private copy of the value. */
  strcpy(Value, UserValue);
  
  /* Remove and free the old value, if any. */
  if ((OldValue = (char *)hash_value_for_key(Table, Key)))
  {
    hash_change_key_value(Table, (void *)Key, (void *)Value);
    free(OldValue);
  }
  else
  {
    /* Insert the new value in the hash table. */
    NewKey = malloc(strlen(Key)+1);
    strcpy(NewKey, Key);
    hash_add(&Table, (void *)NewKey, (void *)Value);
  }
}


/***************
 * FUNCTION:     Cleanup
 * PURPOSE:      Remove illegal characters from a string
 * ARGUMENTS:    char *String : The string to be cleaned up
 * ALGORITHM:    copy the string onto itself, skipping illegal characters
 *               remove trailing whitespaces as well (NB: The string is supposed NOT to
 *               start with ' ' or '\t'
 * RETURN:       void 
 * SIDE EFFECT:  The string might be shortened
 * CALLS:        None
 * CALLED BY:    SETUP_ReadSetup
 ***************/
void Cleanup(char *String)
{
  long i, j;
  
  i = j = 0;
  
  assert(String);
  
  /* Get rid of illegal chars. */
  while (*(String+i))
    if (!strchr(ILLEGAL_CHARS, *(String+i)))
      *(String+j++) = *(String+i++);
    else
      i++;
  *(String+j) = 0;
  
  /* Get rid of trailing whitespaces. */
  j--;
  while(strchr(WHITE_SPACE, *(String+j)))
    *(String+j--) = 0;
  
  return;
}
